// ****************** MUST INCLUDE LINES BELOW*****************************
#include <lib6302.h>
// Init library and set pwm freq, DAC resolution, ADC resolution, and teensy 3.6-3.2 bit rate.
// PWM frequency affects DAC resolution, higher frequency, lower resolution.
// For 8-bit PWM: Teensy 3.6-234375, Teensy 3.2-187500!!!1
// Divide by 2,4,8 or 16 for 9,10,11, or 12 bits.
// WARNING 12 bit pwm is in audio range for young listeners.
#define MYSERIAL Serial
lib6302 my6302(234375/8,12,16,&Serial);
// Set up the analog-to-digital library, with a buffersize
#define sampleBufN 500
LIB6302_ADC(sampleBufN+100);  // ADC sample buffer of 100+sampleBufN
// ****************** MUST INCLUDE LINES ABOVE*******1**********************

// Constants for Users to Modify 
// ************************************************************
#define PLOTALL 0  // 0:theta,thetad 1:add motorcmd, -1:see code
#define ANALOGPARAMS 0  // 1:read Kp and Kd from analog
#define angleOffsetL 0.0  // Adjust if measured angle is not zero when horizontal. 
#define angleOffsetR 0.04 // Adjust if measured angle is not zero when horizontal. 

#define togglePeriod 2.5 // seconds between +/- toggling of desired value.
#define highThetaL 0.3 // Toggle angle in radians high (0.5 is max)
#define lowThetaL -0.3 // Toggle angle in radians low (-0.5 is min)
#define highThetaR 0.4 // Toggle angle in radians high (0.5 is max)
#define lowThetaR -0.4 // Toggle angle in radians low (-0.5 is min)
#define KPMAX 10 // Maximum proportional gain.
#define KDMAX 2.0 // Maximum difference gain 
#define nomCmdL 0.55 // Adjust to zero steady-state error
#define nomCmdR 0.0 // Adjust to zero steady-state error
#define mBack 5
#define plotIncrement 10
#define controllerUpdatePeriod 1000 // Update period in microseconds

// ************************************************************************
// End of Constants and code for Users to Modify 
// SEE END OF FILE FOR USER-ALTERABLE CONTROL LOOP!!

// *******SETTING UP THE MONITOR DISPLAY (SLIDERS AND PLOTS). *************
// Sliders
String sliders = "S~Pause~0~1~S~Kp~0~10~S~Kd~0~5~";
#define PA 0
#define KP 1
#define KD 2

// Plots: USE SPACES BETWEEN LABELS FOR MULTIPLOTS!!
String measDesL = "P~ThetaL_d ThetaL_a~-1~1~500~2~";
String measDesR = "P~ThetaR_d ThetaR_a~-1~1~500~2~";
String ddt = "P~delErrL/Dt delErrR/dt~-1~1~500~2~";
String Cmd = "P~cmdL cmdR~-0~5~500~2~";
String monPlots = measDesL + measDesR + ddt + Cmd;
#define PLOT_LD 0
#define PLOT_LA 1
#define PLOT_RD 2
#define PLOT_RA 3
#define PLOT_DDTL 4
#define PLOT_DDTR 5
#define PLOT_CMDL 6
#define PLOT_CMDR 7

// Configuration string with surrounding limits.
String config_message = sliders + monPlots;

// Control loop setup
// ************************************************************************

// Motor PWM and Sensor defines.
elapsedMicros loopTimer = 0;  // Loop runs at least 10xsecond
float mDeltaT = 0.0;
float headRoom;

void setup() {
  MYSERIAL.begin(1000000);
  my6302.setup(config_message, togglePeriod);
  my6302.startPWM(true); // start parallel drive pwm, A+B in sync.
  my6302.setupADC(16,true);
  loopTimer = 0;
  mDeltaT = mBack*float(controllerUpdatePeriod)*1.0e-6;
  headRoom = controllerUpdatePeriod;
}

#define zeroParamThresh 0.005
#define scaleParam 1.005
float analog2param(int pinA) {
  float retval = my6302.adcGetAvg(pinA) - zeroParamThresh;
  retval = scaleParam * max(retval, 0.0);
  return(retval);
}

//********************** Main USER ALTERABLE Control Loop **********************
//**************************************************************************
int loopCounter = 0;
float pastErrorR[mBack+1];  // Stores past errors for deriv approx
float pastErrorL[mBack+1];

void loop() {  

  // Initialize: check for inputs, flip toggle, etc.
  my6302.startloop();

  // Get parameters.
  float pauseVal= my6302.getSlider(PA);
  float Kp = my6302.getSlider(KP);
  float Kd = my6302.getSlider(KD);
  if(ANALOGPARAMS > 0) { // Get Kp and Kd by reading potentiometers
    Kp = KPMAX*analog2param(A3);
    Kd = KDMAX*analog2param(A2);
  }
  // Headroom is faction of period left until next sensor read.
  float headRoomThis = float(controllerUpdatePeriod - int(loopTimer));
  headRoom = headRoomThis;
  
  // Wait exactly one deltaT since last sensor read.
  while (int(loopTimer) < controllerUpdatePeriod) {};
  loopTimer = 0;
  
  // Read the sensors
  float measR = 2*PI*(0.5 - my6302.adcGetAvg(A1) + angleOffsetR);
  float measL = 2*PI*(0.5 - my6302.adcGetAvg(A0) + angleOffsetL);
  
  // Possibly toggle the desired value.
  float flip = my6302.getFlip();
  float desiredL = highThetaL;
  float desiredR = highThetaR;
  if(flip > 0) {
    desiredL = lowThetaL;
    desiredR = lowThetaR;
  }
  
  // Compute the errors
  float errorL = (desiredL - measL);
  float errorR = (desiredR - measR);
  
  // Save backward angles for delta/deltaT calculation.
  for (int i = mBack; i > 0; i--) pastErrorR[i] = pastErrorR[i-1];
  pastErrorR[0] = errorR;
  for (int i = mBack; i > 0; i--) pastErrorL[i] = pastErrorL[i-1];
  pastErrorL[0] = errorL;

  // Compute (error - oldError)/deltaT.
  float delDtL = (errorL -pastErrorL[mBack])/mDeltaT;
  float delDtR = (errorR -pastErrorR[mBack])/mDeltaT;
  
  // Compute the motor cmd = nominal++Kp*Err.
  float dML = Kp*errorL + Kd*delDtL;
  float dMR = Kp*errorR + Kd*delDtR;
  
  // Write motor command to H-bridge, returns clipped value.
  float rCmd = nomCmdR + dMR;
  float lCmd = nomCmdL + dML;
  float cmdClipL = my6302.hbridgeWrite(lCmd - rCmd, pwmB1);
  float cmdClipR = my6302.hbridgeWrite(lCmd + rCmd, pwmB2_T);
  
  if(loopCounter <  plotIncrement) loopCounter++;
  else {
    loopCounter = 1;
    if(pauseVal == 0) {
      // Send data to monitor.
      float loopStatus[8];
      loopStatus[PLOT_LD] = desiredL;
      loopStatus[PLOT_LA] = measL;
      loopStatus[PLOT_RD] = desiredR;
      loopStatus[PLOT_RA] = measR;
      loopStatus[PLOT_DDTL] = min(max(-2,delDtL/10.0),2);
      loopStatus[PLOT_DDTR] = min(max(-2,delDtR/10.0),2);
      loopStatus[PLOT_CMDL] = cmdClipL;
      loopStatus[PLOT_CMDR] = cmdClipR;
      my6302.sendStatus(8,loopStatus);  
    }
  }
}




  


  
